#ifndef _STRETCHER_HPP_
#define _STRETCHER_HPP_

#if defined SSE && defined __SSE__
	#include <immintrin.h>
#endif

#include <opencv2/core/core.hpp>

namespace histodet {

/**
 * This class does fast histogram stretching
 */
class Stretcher
{
 private:
	int hist_nbits;
	int hist_size;
	int stretch_nbits;
	int stretch_size;
	int step_nbits;
	int min_support;
	bool normalize;
	cv::Mat hist_stretched;
	cv::Mat hist_k, hist_k1;
	int *stretch_k, *stretch_k1;
	float *stretch_x, *stretch_x1;
	const int offset_k1, offset_x, offset_x1;
	
	/**
	For performance reasons (we want x[i] and x1[i] consecutive to avoid having to
	reorganize them to vectorize operations, and we want the whole lut continuous
	in memory for optimal cache use), the data structure is the following one:
	
	struct Stretch { int k[hist_size]; int k1[hist_size]; float x[hist_size]; float x1[hist_size]; }
	Stretch lut[hist_size][hist_size];
	
	But as hist_size is not known at compile time and with the performance constraints
	explained above, we have to hide it
	*/
	void *lut;

 private:
	inline void set_stretches(int a, int b)
	{
		stretch_k = ((int*)lut) + (a << step_nbits) + (b << stretch_nbits);
		stretch_k1 = stretch_k + offset_k1;
		stretch_x  = (float*)(stretch_k + offset_x);
		stretch_x1 = (float*)(stretch_k + offset_x1);
	}
 public:
	/**
	 * Constructor
	 * @param hist_nbits number of bits of histogram size
	 * @param min_support minimum width (in number of buckets) of the non-null values of the histogram to stretch
	 * @param normalize whether the values should be stretched after normalizing. In theory it should, but
	 * in practice sometimes it works better without.
	 */
	Stretcher(int hist_nbits, int min_support, bool normalize);
	/**
	 * Destructor
	 */
	~Stretcher();
	/**
	 * Compute the first and last bucket index of the histogram that are non null
	 * @param hist the histogram
	 * @param min_val the value under which a bucket is considered null
	 * @param[out] a the first value
	 * @param[out] b the last value
	 */
	inline void computeSupport(cv::Mat &hist, float min_val, int &a, int &b);
	/**
	 * Stretch the histogram
	 * @param hist the histogram
	 * @param min_val the value under which a bucker is considered null
	 * @param[out] zap whether the support was too small and the histogram was not stretched
	 * @return the stretched histogram
	 */
	inline cv::Mat& stretch(cv::Mat &hist, float min_val, bool &zap);

};



///*********************************************************************************************************************************************
/// Inlined functions
///*********************************************************************************************************************************************


void Stretcher::computeSupport(cv::Mat &hist, float min_val, int &a, int &b)
{
	float *h = (float*)(hist.data);
	a = 0; while (h[a] < min_val) ++a;
	b = hist_size-1; while (h[b] < min_val) --b;
}

cv::Mat& Stretcher::stretch(cv::Mat &hist, float min_val, bool &zap)
{
	// compute bounds a,b
	int a, b;
	computeSupport(hist, min_val, a, b);
	if (b-a < min_support) { zap = true; return hist_stretched; }
	
	// now stretch
	set_stretches(a, b);

#if 0 && defined SSE && defined __SSE__
	for(int i = 0; i < hist_size; ++i)
	{
		((float*)hist_k .data)[i] = h[stretch_k [i]];
		((float*)hist_k1.data)[i] = h[stretch_k1[i]];
	}
// 	for(int i = 0; i < hist_size; ++i)
// 		((float*)(hist_stretched.data))[i] = ((float*)hist_k .data)[i] * stretch_x[i] + ((float*)hist_k1.data)[i] * stretch_x1[i];
	for(int i = 0; i < hist_size; i+=4)
		*((__m128*)((float*)hist_stretched.data+i)) = _mm_add_ps(
			_mm_mul_ps(*((__m128*)((float*)hist_k .data+i)), *((__m128*)(stretch_x +i))),
			_mm_mul_ps(*((__m128*)((float*)hist_k1.data+i)), *((__m128*)(stretch_x1+i))));

	// another solution (if we precompute correctly shift and flag), but it is anyway even slower!
// 	for(int i = 0; i < hist_size; i+=4)
// 	{
// 		*((__m128*)((float*)hist_stretched.data+i)) = _mm_add_ps(
// 			_mm_mul_ps(_mm_shuffle_ps(*((__m128*)stretch_k ) , *((__m128*)stretch_k ), 0x00), *((__m128*)(stretch_x +i))),
// 			_mm_mul_ps(_mm_shuffle_ps(*((__m128*)stretch_k1), *((__m128*)stretch_k1), 0x00), *((__m128*)(stretch_x1+i))));
// 	}
#else
	float *h = (float*)(hist.data);
	for(int i = 0; i < hist_size; ++i)
		((float*)(hist_stretched.data))[i] = h[stretch_k[i]] * stretch_x[i] + h[stretch_k1[i]] * stretch_x1[i];
#endif

	zap = false;
	return hist_stretched;
}



} // namespace histodet





#endif






